#define RESEARCH03_DESCRIPTION "         ."

#include "Result.mqh"

#define _CS(A) ((!::IsStopped()) && (A))
#define DAY (24 * 3600)

class RESEARCH03
{
private:
  PRICE Prices[];
  PRICE PricesOOS[];

  int MaxLength1;
  int MaxLength2;

  const int DayBars;

  static datetime TimeToDay( const datetime time )
  {
    return(time / DAY * DAY);
  }


  static int GetAmountDays( const PRICE &Prices[] )
  {
    int Res = 0;
    datetime PrevDay = 0;

    for (int i = ::ArraySize(Prices) - 1; i >= 0; i--)
    {
      const datetime Day = TimeToDay(Prices[i].time);

      if (Day != PrevDay)
      {
        Res++;

        PrevDay = Day;
      }
    }

    return(Res);
  }


  static int GetMaxBars( const datetime FromDate, const datetime ToDate, const ENUM_TIMEFRAMES period = PERIOD_CURRENT )
  {
    int Res = 0;
    const int Step = ::PeriodSeconds(period);

    for (datetime time = FromDate; time < ToDate; time += Step)
    {
      const int DayOfWeek = TimeDayOfWeek(time);

      if (DayOfWeek && DayOfWeek < 6)
        Res++;
    }

    return(Res);
  }

  static int PriceToInteger( const double Price, const double point )
  {
    return((int)(Price / point + 0.1));
  }


  static int FillPrices( const string Symb, PRICE &Prices[], datetime FromDate, datetime ToDate, const ENUM_TIMEFRAMES period = PERIOD_CURRENT )
  {
    MqlRates Rates[];

    const int Step = ::PeriodSeconds(period);

    FromDate = RESEARCH03::TimeToDay(FromDate);
    ToDate = RESEARCH03::TimeToDay(ToDate);

    int Amount = 0;
    const int Size = ::CopyRates(Symb, period, FromDate, ToDate, Rates);

    if (Size)
    {
      const double point = ::SymbolInfoDouble(Symb, SYMBOL_POINT);

      FromDate = RESEARCH03::TimeToDay(Rates[0].time);
      ToDate = RESEARCH03::TimeToDay(Rates[Size - 1].time + DAY);

      ::ArrayResize(Prices, RESEARCH03::GetMaxBars(FromDate, ToDate, period));

      int i = 0;

      for (datetime time = FromDate; (time < ToDate) && (i < Size); time += Step)
      {
        const int DayOfWeek = ::TimeDayOfWeek(time);

        if (DayOfWeek && DayOfWeek < 6)
        {
          Prices[Amount].time = time;

          if (time == Rates[i].time)
            Prices[Amount++].Price = RESEARCH03::PriceToInteger(Rates[i++].open, point);
          else if (time > Rates[i].time)
            Prices[Amount++].Price = RESEARCH03::PriceToInteger(Rates[i++].close, point);
          else
          {
            Prices[Amount].Price = Amount ? Prices[Amount - 1].Price : RESEARCH03::PriceToInteger(Rates[i].open, point);

            Amount++;
          }
        }
      }
    }

    return(::ArrayResize(Prices, Amount));
  }

public:
  RESEARCH03( const string sSymb, const datetime iFromDate, const ENUM_TIMEFRAMES ePeriod = PERIOD_CURRENT,
              const int iMaxLength1 = 0, const int iMaxLength2 = 0 ) : DayBars(DAY / ::PeriodSeconds(ePeriod)),
                                                                       MaxLength1(iMaxLength1), MaxLength2(iMaxLength2)
  {
    const int Size = RESEARCH03::FillPrices(sSymb, this.Prices, iFromDate, ::TimeCurrent(), ePeriod);

    if (Size > 0)
      RESEARCH03::FillPrices(sSymb, this.PricesOOS, (this.Prices[0].time << 1) - this.Prices[Size - 1].time, this.Prices[0].time, ePeriod);
  }

  void Run() const
  {
    const int Size = ::ArraySize(this.Prices);

    if (Size > 0)
    {
      const int Days = RESEARCH03::GetAmountDays(this.Prices);
      const int DaysOOS = RESEARCH03::GetAmountDays(this.PricesOOS);

      Print("Research's Days = " + (string)Days + ": " +
            ::TimeToString(this.Prices[0].time, TIME_DATE) + "-" +
            ::TimeToString(this.Prices[Size - 1].time, TIME_DATE) +
            ". OOS's Days = " + (string)DaysOOS + ": " +
            ::TimeToString(this.PricesOOS[0].time, TIME_DATE) + "-" +
            ::TimeToString(this.PricesOOS[::ArraySize(this.PricesOOS) - 1].time, TIME_DATE));

      RESULT MaxRes = 0;

      const int Length1 = this.MaxLength1 ? (int)::MathMin(DayBars, this.MaxLength1) : this.DayBars;
      const int Length2 = this.MaxLength2 ? (int)::MathMin(DayBars, this.MaxLength2) : this.DayBars;

      const datetime StartTime = ::TimeLocal();

      for (int Begin1 = 0; _CS(Begin1 < DayBars); Begin1++)
      {
        const double Procent = (Begin1 + 1.0) / DayBars;

        if (Procent)
        {
          const datetime NowTime = ::TimeLocal();

          const string Str = "Done " + ::DoubleToString(Procent * 100, 2) + "%" + ", " +
                             ::TimeToString(NowTime - StartTime, TIME_SECONDS) + " / " +
                             ::TimeToString((datetime)((NowTime - StartTime) / Procent), TIME_SECONDS) +
                             "\nBegin1 = " + ::TimeToString(this.Prices[Begin1].time, TIME_MINUTES);

          ::Comment(Str);
        }

        for (int End1 = Begin1 + 1, StopEnd1 = Begin1 + Length1; _CS(End1 < StopEnd1); End1++)
          for (int Begin2 = End1, StopBegin2 = End1 + DayBars; Begin2 < StopBegin2; Begin2++)
            for (int End2 = Begin2 + 1, StopEnd2 = Begin2 + Length2; End2 < StopEnd2; End2++)
            {
              int Res = 0;

              for (int _Begin1 = Begin1, _End1 = End1, _Begin2 = Begin2, _End2 = End2; _End2 < Size;
                   _Begin1 += DayBars, _End1 += DayBars, _Begin2 += DayBars, _End2 += DayBars)
              {
                const double Diff = (this.Prices[_End1].Price - this.Prices[_Begin1].Price) * (this.Prices[_End2].Price - this.Prices[_Begin2].Price);

                if (Diff > 0)
                  Res++;
                else if (Diff < 0)
                  Res--;
              }

              if (MaxRes <= Res)
              {
                MaxRes = Res;
                MaxRes.SetIntervals(Begin1, End1, Begin2, End2);

                const int ResOOS = MaxRes.Get(this.PricesOOS);

                ::Print(::DoubleToString(100.0 * (MaxRes.Value + Days) / (Days << 1), 1) + "%, " + MaxRes.ToString(this.Prices) +
                        ", OOS:: " + (string)ResOOS + ", " + (string)DaysOOS + " days - " + DoubleToString(100.0 * (ResOOS + DaysOOS) / (DaysOOS << 1), 1) + "%");
              }
            }
        }

      ::Comment("");
    }
  }
};

#undef DAY
#undef _CS